@@ -1,6 +1,7 @@
-use std::collections::HashSet;
+use std::collections::{hash_map::DefaultHasher, HashSet, HashMap};
 use std::error::Error;
 use std::fs::{self, File};
+use std::hash::{Hash, Hasher};
 use std::io;
 use std::io::prelude::*;
 use std::path::{Path, PathBuf};
@@ -23,7 +24,7 @@ use crate::consts;
 mod error;
 use error::Error as FDroidError;
 
-async fn retrieve_index_or_exit() -> Value {
+async fn retrieve_index_or_exit(options: &HashMap<&str, &str>) -> Value {
     let temp_dir = match tempdir() {
         Ok(temp_dir) => temp_dir,
         Err(_) => {
@@ -31,6 +32,24 @@ async fn retrieve_index_or_exit() -> Value {
             std::process::exit(1);
         }
     };
+    let mut custom_repo = false;
+    let mut repo = consts::FDROID_REPO.to_string();
+    let mut fingerprint = Vec::from(consts::FDROID_INDEX_FINGERPRINT);
+    if let Some(full_repo_option) = options.get("repo") {
+        custom_repo = true;
+        if let Some((repo_option, fingerprint_option)) = full_repo_option.split_once("?fingerprint=") {
+            fingerprint = match hex::decode(fingerprint_option) {
+                Ok(hex_fingerprint) => hex_fingerprint,
+                Err(_) => {
+                    println!("Fingerprint must be specified as valid hex. Exiting.");
+                    std::process::exit(1);
+                }
+            };
+            repo = repo_option.to_string();
+        } else {
+            repo = full_repo_option.to_string();
+        }
+    }
     let config_dir = match dirs::config_dir() {
         Some(mut config_dir) => {
             let create_dir = |config_dir: &PathBuf| {
@@ -42,6 +61,14 @@ async fn retrieve_index_or_exit() -> Value {
             create_dir(&config_dir);
             config_dir.push("apkeep");
             create_dir(&config_dir);
+            if custom_repo {
+                config_dir.push("fdroid-custom-repos");
+                create_dir(&config_dir);
+                let mut s = DefaultHasher::new();
+                repo.hash(&mut s);
+                config_dir.push(format!("{}", s.finish()));
+                create_dir(&config_dir);
+            }
             config_dir
         },
         None => {
@@ -64,8 +91,9 @@ async fn retrieve_index_or_exit() -> Value {
     };
 
     let http_client = reqwest::Client::new();
+    let fdroid_index_url = format!("{}/index-v1.jar", repo);
     let index_response = http_client
-        .head(consts::FDROID_INDEX_URL)
+        .head(fdroid_index_url)
         .send().await.unwrap();
 
     let etag = if index_response.headers().contains_key("ETag") {
@@ -86,8 +114,12 @@ async fn retrieve_index_or_exit() -> Value {
         };
         serde_json::from_str(&index).unwrap()
     } else {
-        let files = download_and_extract_index_to_tempdir(&temp_dir).await;
-        match verify_and_return_index(&temp_dir, &files) {
+        let files = download_and_extract_index_to_tempdir(&temp_dir, &repo).await;
+        let verify_index = match options.get("verify-index") {
+            Some(&"false") => false,
+            _ => true,
+        };
+        match verify_and_return_index(&temp_dir, &files, &fingerprint, verify_index) {
             Ok(index) => {
                 match serde_json::from_str(&index) {
                     Ok(index_value) => {
@@ -107,7 +139,8 @@ async fn retrieve_index_or_exit() -> Value {
                     }
                 }
             },
-            Err(_) => {
+            Err(err) => {
+                println!("{}", err);
                 println!("Could not verify F-Droid package index. Exiting.");
                 std::process::exit(1);
             },
@@ -120,8 +153,9 @@ pub async fn download_apps(
     parallel: usize,
     sleep_duration: u64,
     outpath: &Path,
+    options: HashMap<&str, &str>,
 ) {
-    let index = retrieve_index_or_exit().await;
+    let index = retrieve_index_or_exit(&options).await;
 
     let (fdroid_apps, repo_address) = match parse_json_for_download_information(index, apps) {
         Ok((fdroid_apps, repo_address)) => (fdroid_apps, repo_address),
@@ -258,8 +292,8 @@ fn parse_json_for_download_information(index: Value, apps: Vec<(String, Option<S
     Ok((fdroid_apps, repo_address.to_string()))
 }
 
-pub async fn list_versions(apps: Vec<(String, Option<String>)>) {
-    let index = retrieve_index_or_exit().await;
+pub async fn list_versions(apps: Vec<(String, Option<String>)>, options: HashMap<&str, &str>) {
+    let index = retrieve_index_or_exit(&options).await;
     if parse_json_display_versions(index, apps).is_err() {
         println!("Could not parse JSON of F-Droid package index. Exiting.");
     };
@@ -303,8 +337,7 @@ fn parse_json_display_versions(index: Value, apps: Vec<(String, Option<String>)>
     Ok(())
 }
 
-fn verify_and_return_index(dir: &TempDir, files: &[String]) -> Result<String, Box<dyn Error>> {
-    println!("Verifying...");
+fn verify_and_return_index(dir: &TempDir, files: &[String], fingerprint: &[u8], verify_index: bool) -> Result<String, Box<dyn Error>> {
     let re = Regex::new(consts::FDROID_SIGNATURE_BLOCK_FILE_REGEX).unwrap();
     let cert_file = {
         let mut cert_files = vec![];
@@ -327,22 +360,27 @@ fn verify_and_return_index(dir: &TempDir, files: &[String]) -> Result<String, Bo
         signed_file
     };
 
-    let signed_file_data = fs::read(signed_file)?;
+    let signed_content = fs::read(signed_file)?;
 
-    {
-        let (cert, signer_info) = get_certificate_and_signer_info_from_signature_block_file(cert_file)?;
-        cert.verify_signed_data(signed_file_data.clone(), signer_info.signature())?;
+    if verify_index {
+        println!("Verifying...");
+        let signed_data = get_signed_data_from_cert_file(cert_file)?;
+        let signer_info = signed_data.signers().next().unwrap();
+        signer_info.verify_signature_with_signed_data_and_content(
+            &signed_data,
+            &signed_content)?;
+        let cert = signed_data.certificates().next().unwrap();
         let x509 = X509::from_der(&cert.encode_ber()?)?;
-        let fingerprint = x509.digest(MessageDigest::from_name("sha256").unwrap())?;
-        if fingerprint.as_ref() != consts::FDROID_INDEX_FINGERPRINT {
+        let cert_fingerprint = x509.digest(MessageDigest::from_name("sha256").unwrap())?;
+        if cert_fingerprint.as_ref() != fingerprint {
             return Err(Box::new(SimpleError::new("Fingerprint of the key contained in the F-Droid repository index does not match the expected fingerprint.")))
         };
     }
 
-    let signed_file_string = std::str::from_utf8(&signed_file_data)?;
+    let signed_file_string = std::str::from_utf8(&signed_content)?;
     let manifest_file = dir.path().join("META-INF").join("MANIFEST.MF");
     let manifest_file_data = fs::read(manifest_file)?;
-    {
+    if verify_index {
         let signed_file_regex = Regex::new(r"\r\nSHA1-Digest-Manifest: (.*)\r\n").unwrap();
         let signed_file_manifest_sha1sum = base64::decode(match signed_file_regex.captures(signed_file_string) {
             Some(caps) if caps.len() >= 2 => caps.get(1).unwrap().as_str(),
@@ -361,7 +399,7 @@ fn verify_and_return_index(dir: &TempDir, files: &[String]) -> Result<String, Bo
     let manifest_file_string = std::str::from_utf8(&manifest_file_data)?;
     let index_file = dir.path().join("index-v1.json");
     let index_file_data = fs::read(index_file)?;
-    {
+    if verify_index {
         let manifest_file_regex = Regex::new(r"\r\nName: index-v1\.json\r\nSHA1-Digest: (.*)\r\n").unwrap();
         let manifest_file_index_sha1sum = base64::decode(match manifest_file_regex.captures(manifest_file_string) {
             Some(caps) if caps.len() >= 2 => caps.get(1).unwrap().as_str(),
@@ -380,7 +418,7 @@ fn verify_and_return_index(dir: &TempDir, files: &[String]) -> Result<String, Bo
     Ok(String::from(std::str::from_utf8(&index_file_data)?))
 }
 
-fn get_certificate_and_signer_info_from_signature_block_file(signature_block_file: PathBuf) -> Result<(CapturedX509Certificate, SignerInfo), Box<dyn Error>> {
+fn get_signed_data_from_cert_file(signature_block_file: PathBuf) -> Result<SignedData, Box<dyn Error>> {
     let bytes = fs::read(signature_block_file).unwrap();
     match SignedData::parse_ber(&bytes) {
         Ok(signed_data) => {
@@ -398,16 +436,19 @@ fn get_certificate_and_signer_info_from_signature_block_file(signature_block_fil
             if signatories.is_empty() {
                 return Err(Box::new(SimpleError::new("No signatories provided.")));
             }
-            Ok((certificates[0].clone(), signatories[0].clone()))
+            Ok(signed_data)
         },
-        Err(err) => Err(Box::new(err)),
+        Err(err) => {
+            Err(Box::new(err))
+        }
     }
 }
 
-async fn download_and_extract_index_to_tempdir(dir: &TempDir) -> Vec<String> {
+async fn download_and_extract_index_to_tempdir(dir: &TempDir, repo: &str) -> Vec<String> {
     println!("Downloading F-Droid package repository...");
     let mut files = vec![];
-    match tokio_dl_stream_to_disk::download(consts::FDROID_INDEX_URL, dir.path(), "index.zip").await {
+    let fdroid_index_url = format!("{}/index-v1.jar", repo);
+    match tokio_dl_stream_to_disk::download(fdroid_index_url, dir.path(), "index.zip".to_string()).await {
         Ok(_) => {
             println!("Package repository downloaded successfully!\nExtracting...");
             let file = fs::File::open(dir.path().join("index.zip")).unwrap();
